/********************************************************
 *  ██████╗  ██████╗████████╗██╗
 * ██╔════╝ ██╔════╝╚══██╔══╝██║
 * ██║  ███╗██║        ██║   ██║
 * ██║   ██║██║        ██║   ██║
 * ╚██████╔╝╚██████╗   ██║   ███████╗
 *  ╚═════╝  ╚═════╝   ╚═╝   ╚══════╝
 * Geophysical Computational Tools & Library (GCTL)
 *
 * Copyright (c) 2022  Yi Zhang (yizhang-geo@zju.edu.cn)
 *
 * GCTL is distributed under a dual licensing scheme. You can redistribute 
 * it and/or modify it under the terms of the GNU Lesser General Public 
 * License as published by the Free Software Foundation, either version 2 
 * of the License, or (at your option) any later version. You should have 
 * received a copy of the GNU Lesser General Public License along with this 
 * program. If not, see <http://www.gnu.org/licenses/>.
 * 
 * If the terms and conditions of the LGPL v.2. would prevent you from using 
 * the GCTL, please consider the option to obtain a commercial license for a 
 * fee. These licenses are offered by the GCTL's original author. As a rule, 
 * licenses are provided "as-is", unlimited in time for a one time fee. Please 
 * send corresponding requests to: yizhang-geo@zju.edu.cn. Please do not forget 
 * to include some description of your company and the realm of its activities. 
 * Also add information on how to contact you by electronic and paper mail.
 ******************************************************/

#include "olayer_rmse.h"

gctl::rmse::rmse() {}

gctl::rmse::~rmse() {}

void gctl::rmse::evaluation(const matrix<double> &prev_layer_data, const matrix<double> &target)
{
    int nobs = prev_layer_data.col_size();
    int nvar = prev_layer_data.row_size();

    if (target.col_size() != nobs || target.row_size() != nvar)
    {
        throw std::invalid_argument("[gctl::rmse] Target data have incorrect dimension.");
    }
    
    // Compute the derivative of the input of this layer
    // L = 0.5 * ||yhat - y||^2
    // in = yhat
    // d(L) / d(in) = yhat - y
    der_in_.resize(nvar, nobs);

    int i, j;
#pragma omp parallel for private (i, j) schedule(guided)
    for (i = 0; i < nvar; i++)
    {
        for (j = 0; j < nobs; j++)
        {
            der_in_[i][j] = prev_layer_data[i][j] - target[i][j];
        }
    }
    return;
}

double gctl::rmse::loss_value() const
{
    // L = 0.5 * ||yhat - y||^2
    double n = der_in_.norm(L2);
    return 0.5*power2(n)/der_in_.col_size();
}

std::string gctl::rmse::get_output_name() const
{
    return "RegressionMSE";
}

gctl::olayer_type_e gctl::rmse::get_output_type() const
{
    return RegressionMSE;
}